// vim: et:sw=4:ts=4
// C++-code to realise the MATLAB-Papaya-interface
// bug reports, questions and comments to Jenny Wagner (j.wagner@uni-heidelberg.de)
// first include might require to adapt path to mex.h in the MATLAB distribution that is installed:
// adapt the following line if necessary: <path to MATLAB distribution>/extern/include/mex.h (most probably)
#include "/usr/local/MATLAB/R2015a/extern/include/mex.h" 
#include <iostream>
#include <fstream>
#include <getopt_pp_standalone.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <string>
#include <errno.h>
#include "util.h"
#include "minkval.h"
using namespace GetOpt;

typedef AbstractMinkowskiFunctional **func_iterator;

// recursively create directories for the output files
static
void prefix_mkdir (std::string prefix)
{
    std::string::size_type p = prefix.find_last_of ("/");
    if (p == prefix.npos) return;
    prefix.assign (prefix, 0u, p);
    prefix_mkdir (prefix);
    errno = 0;
    std::cerr << "[papaya] mkdir " << prefix << "\n";
    mkdir (prefix.c_str (), 0777);
    if (errno && errno != EEXIST)
        perror ("mkdir");
}

static void calculate_functional (AbstractMinkowskiFunctional *p,
                                  const Boundary &b) {
    Boundary::contour_iterator cit;
    for (cit = b.contours_begin (); cit != b.contours_end (); ++cit)
        p->add_contour (b, b.edges_begin (cit), b.edges_end (cit));
}

static void set_refvert_com (func_iterator begin, func_iterator end,
                             const Boundary &b, int num_labels) {
    VectorMinkowskiFunctional *w010 = create_w010 ();
    ScalarMinkowskiFunctional *w000 = create_w000 ();
    w010->global_ref_vertex (vec_t (0., 0.));
    calculate_functional (w010, b);
    calculate_functional (w000, b);
    for (; begin != end; ++begin) {
        for (int l = 0; l != num_labels; ++l) {
            vec_t refvert = w010->value (l) / w000->value (l);
            (*begin)->ref_vertex (l, refvert);
        }
    }
    delete w010;
    delete w000;
}

static void set_refvert_cos (func_iterator begin, func_iterator end,
                             const Boundary &b, int num_labels) {
    VectorMinkowskiFunctional *w110 = create_w110 ();
    ScalarMinkowskiFunctional *w100 = create_w100 ();
    w110->global_ref_vertex (vec_t (0., 0.));
    calculate_functional (w110, b);
    calculate_functional (w100, b);
    for (; begin != end; ++begin) {
        for (int l = 0; l != num_labels; ++l) {
            vec_t refvert = w110->value (l) / w100->value (l);
            (*begin)->ref_vertex (l, refvert);
        }
    }
    delete w110;
    delete w100;
}

static void set_refvert_coc (func_iterator begin, func_iterator end,
                             const Boundary &b, int num_labels) {
    VectorMinkowskiFunctional *w210 = create_w210 ();
    ScalarMinkowskiFunctional *w200 = create_w200 ();
    w210->global_ref_vertex (vec_t (0., 0.));
    calculate_functional (w210, b);
    calculate_functional (w200, b);
    for (; begin != end; ++begin) {
        for (int l = 0; l != num_labels; ++l) {
            if (fabs (w200->value (l) / W2_NORMALIZATION) < .95*M_PI)
            {
                std::cerr << "error: some labels have vanishing total curvature.\n"
                             "the _coc reference vertex does not exist in this case.\n"
                             "you probably want to use point_of_reference = contour_com or component_com instead.\n";
                exit (1);
            }
            vec_t refvert = w210->value (l) / w200->value (l);
            (*begin)->ref_vertex (l, refvert);
        }
    }
    delete w210;
    delete w200;
}

static void set_refvert_origin (func_iterator begin, func_iterator end,
                                int num_labels) {
    vec_t refvert = vec_t (0., 0.);
    for (; begin != end; ++begin) {
        for (int l = 0; l != num_labels; ++l) {
            (*begin)->ref_vertex (l, refvert);
        }
    }
}

// function required to take image array from MATLAB and convert it into a Pixmap instance
static void load_picture(Pixmap *p, double* matlab_matrix, int rows, int cols) {
    
    // initialise Pixmap
    assert (p);
    p->resize(cols,rows);
    
    // copy values: size1 = cols = x-values, size2 = rows = y-values
    for (int j = 0; j != p->size2 (); ++j)
    for (int i = 0; i != p->size1 (); ++i) {
            double tmp = matlab_matrix[rows-1-j+rows*(i)];
            (*p)(i,j) = Pixmap::val_t (tmp);
        }
    
}

// the interface function to MATLAB
void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[]) {    

    // initialse variable to set number of entries in the output array
    int num_results = 0;
    
    // retrieve arguments of MATLAB call function: 
    // integer number that defines the Minkowski functional to be computed (see MATLAB for reference)
    int mf_name = mxGetScalar(prhs[0]);
    // integer to define origin (picture origin, centre of mass, centre of perimeter or centre of curvature)
    int mf_origin = mxGetScalar(prhs[1]); 
    // integer to define method to be used (by contour, by component)
    // by domain: not necessary as domains can be defined in the MATLAB function
    // (implemented as "int" such that it can be extended in future versions)
    int mf_method = mxGetScalar(prhs[2]); 
    // double to set threshold if image is to be segmented before calculation (white = 1.0, black = 0.0)
    double threshold = mxGetScalar(prhs[3]);
    // invert image
    bool invert_image = mxGetScalar(prhs[4]); 
    // prefer connecting the white object (false/ 0) or the black background phase (true/ 1)
    bool connect_black = mxGetScalar(prhs[5]);
    // is data periodic (true / false)
    bool periodic_data = mxGetScalar(prhs[6]);
    // retrieve number of lines and columns in image and the image itself
    int rows = mxGetM(prhs[7]);
    int cols = mxGetN(prhs[7]);
    double* picture = mxGetPr(prhs[7]);
    // integer to set precision of output (actually not necessary as MATLAB usually uses doubles)
    int precision = mxGetScalar(prhs[8]);
    // boolean to print contours and labels to file (print = 1, do not print = 0)
    bool dump_lab = mxGetScalar(prhs[9]);
    std::string label_path;
    
    if(dump_lab)
         label_path = mxArrayToString(prhs[10]);
    
    std::cout << "--------------------------------------INPUT--------------------------------" << std::endl;
    std::cout << "parameters: " << std::endl;
    std::cout << "mf_name: " << mf_name << "   mf_origin: " << mf_origin << "   mf_method: " << mf_method << "   threshold: " << threshold << std::endl;
    std::cout << "invert_image: " << invert_image <<  "   connect_black: " << connect_black << "   periodic_data: " << periodic_data << "   precision: " << precision << std::endl;
    std::cout << "size of picture (rows x cols): " << rows << " x " << cols << std::endl;
    if(dump_lab)
        std::cout << "writing out labels to file " << label_path << "labels.out" << std::endl;
    std::cout << "---------------------------------------------------------------------------" << std::endl;

    Boundary b;
    Pixmap p;
    // convert MATLAB picture into Pixmap
    load_picture(&p, picture, rows, cols);
    
    // std::cout << "file read in complete." << std::endl;
    if(invert_image)
        invert (&p);
    
    marching_squares (&b, p, threshold, connect_black, periodic_data);

    // std::cout << "marching squares complete." << std::endl;
    
    assert_sensible_boundary (b);
    
    // std::cout << "assert boundary complete." << std::endl;
    

    // calculate everything by contour or by component
    int num_labels; 
    switch (mf_method){
        case 1: {
            num_labels = label_by_contour_index (&b);
            break;
        }
        case 2: {
            num_labels = label_by_component (&b);
            break;
        }
        default: {
            std::cerr << "integer between 1 - 2 required to specify labels by contour or component." << std::endl;
        }
    
    }
       
    // std::cout << "num_labels complete." << std::endl;
    
    // write out labels if desired 
    if(dump_lab){
        std::string label_file (label_path);
        dump_labels (label_file, b);
    }
    
    // decide from first argument which MF to calculate and set general pointer to that MF
    AbstractMinkowskiFunctional *m_func; 
    switch(mf_name) {
        case 1: {
            ScalarMinkowskiFunctional *w000 = create_w000 (); 
            m_func = w000;
            num_results = 1; 
            break; }
        case 2: {
            ScalarMinkowskiFunctional *w100 = create_w100 (); 
            m_func = w100;
            num_results = 1;
            break; 
        }
        case 3: {
            ScalarMinkowskiFunctional *w200 = create_w200 ();
            m_func = w200;
            num_results = 1;
            break;
            
        }
        case 4: {
            VectorMinkowskiFunctional *w010 = create_w010 ();
            m_func = w010;
            num_results = 2;
            break; 
        }
        case 5: {
            VectorMinkowskiFunctional *w110 = create_w110 ();
            m_func = w110;
            num_results = 2;
            break;
        }
        case 6: {
            VectorMinkowskiFunctional *w210 = create_w210 ();
            m_func = w210;
            num_results = 2;
            break;
        }
        case 7: {
            MatrixMinkowskiFunctional *w020 = create_w020 ();
            m_func = w020;
            num_results = 10; 
            break;
        }
        case 8: {
            MatrixMinkowskiFunctional *w120 = create_w120 ();
            m_func = w120;
            num_results = 10;
            break;
        }
        case 9: {
            MatrixMinkowskiFunctional *w102 = create_w102 ();
            m_func = w102;
            num_results = 10;
            break;
        }
        case 10: {
            MatrixMinkowskiFunctional *w220 = create_w220 ();
            m_func = w220;
            num_results = 10;
            break;
        }
        case 11: {
            MatrixMinkowskiFunctional *w211 = create_w211 ();
            m_func = w211;
            num_results = 10;
            break;
        }
        default: {
            std::cerr << "integer between 1 - 11 required to specify Minkowski functional to be computed." << std::endl;
        }
    
    }
    
    AbstractMinkowskiFunctional* m_func_array[] = {m_func};
    func_iterator func_begin = m_func_array;

    std::cout << "origin set to "; 
    switch(mf_origin) {
        case 1: {
            set_refvert_origin (func_begin, func_begin + 1, num_labels);
            std::cout << "picture origin." << std::endl;
            break;
        }
        case 2: {
            set_refvert_com (func_begin, func_begin + 1, b, num_labels);
            std::cout << "centre of light of image." << std::endl;
            break;
        }
        case 3: {
            set_refvert_cos (func_begin, func_begin + 1, b, num_labels);
            std::cout << "centre of perimeter of image." << std::endl;
            break;
        }
        case 4: {
            set_refvert_coc (func_begin, func_begin + 1, b, num_labels);
            std::cout << "centre of curvature of image." << std::endl;
            break;
        }
        default: {
            std::cerr << "integer between 1 - 4 required to specify origin." << std::endl;
        }
    }
            
    // calculate functional
    calculate_functional (m_func, b);
    
    // std::cout << "functional calculated." << std::endl;
    
    // generate MATLAB output
    plhs[0] = mxCreateDoubleMatrix(num_labels, num_results, mxREAL);
    double* output_ptr = mxGetPr(plhs[0]);
    
    // plot calculated MF to screen
    std::cout << "-------------------------------------OUTPUT--------------------------------" << std::endl;
    switch(mf_name) {
        case 1: {
            std::cout << "w000:"  << std::endl;
            ScalarMinkowskiFunctional *p = dynamic_cast<ScalarMinkowskiFunctional*>(m_func);
            for (int l = 0; l != num_labels; ++l) {
                output_ptr[l] = p->value(l);
                // std::cout << std::setw (19) << l << " " << std::setw (19) << std::setprecision (precision) << p->value (l) << std::endl;
                
            }
            break; 
        }
        case 2: {
            std::cout << "w100:"  << std::endl;
            ScalarMinkowskiFunctional *p = dynamic_cast<ScalarMinkowskiFunctional*>(m_func);
            for (int l = 0; l != num_labels; ++l) {
                output_ptr[l] = p->value(l);
                // std::cout << std::setw (19) << l << " " << std::setw (19) << std::setprecision (precision) << p->value (l) << std::endl;
            };
            break; 
        }
        case 3: {
            std::cout << "w200:"  << std::endl;
            ScalarMinkowskiFunctional *p = dynamic_cast<ScalarMinkowskiFunctional*>(m_func);
            for (int l = 0; l != num_labels; ++l) {
                output_ptr[l] = p->value(l);
                // std::cout << std::setw (19) << l << " " << std::setw (19) << std::setprecision (precision) << p->value (l) << std::endl;
            };
            break;
        }
        case 4: {
            std::cout << "w010:"  << std::endl;
            VectorMinkowskiFunctional *p = dynamic_cast<VectorMinkowskiFunctional*>(m_func);
            for (int l = 0; l != num_labels; ++l) {
                vec_t val = p->value (l);
                output_ptr[l] = val.x();
                output_ptr[l + num_labels] = val.y();
                std::cout << std::setw (19) << l << " " << std::setw (19) << std::setprecision (precision) << val.x () << " " << std::setw (19) << std::setprecision (precision) << val.y () << std::endl;
            };        
            break; 
        }
        case 5: {
            std::cout << "w110:"  << std::endl;
            VectorMinkowskiFunctional *p = dynamic_cast<VectorMinkowskiFunctional*>(m_func);
            for (int l = 0; l != num_labels; ++l) {
                vec_t val = p->value (l);
                output_ptr[l] = val.x();
                output_ptr[l + num_labels] = val.y();
                // std::cout << std::setw (19) << l << " " << std::setw (19) << std::setprecision (precision) << val.x () << " " << std::setw (19) << std::setprecision (precision) << val.y () << std::endl;
            };        
            break; 
        }
        case 6: {
            std::cout << "w210:"  << std::endl;
            VectorMinkowskiFunctional *p = dynamic_cast<VectorMinkowskiFunctional*>(m_func);
            for (int l = 0; l != num_labels; ++l) {
                vec_t val = (p)->value (l);
                output_ptr[l] = val.x();
                output_ptr[l + num_labels] = val.y();
                // std::cout << std::setw (19) << l;
                // std::cout << " " << std::setw (19) << std::setprecision (precision) << val.x ();
                // std::cout << " " << std::setw (19) << std::setprecision (precision) << val.y () << std::endl;
            };        
            break; 
        }
        case 7: {
            std::cout << "w020:"  << std::endl;
            MatrixMinkowskiFunctional *p = dynamic_cast<MatrixMinkowskiFunctional*>(m_func);
            for (int l = 0; l != num_labels; ++l) {
                mat_t val = p->value (l);
                output_ptr[l] = val(0,0);
                output_ptr[l + num_labels] = val(0,1);
                output_ptr[l + 2*num_labels] = val(1,0);
                output_ptr[l + 3*num_labels] = val(1,1);
                // std::cout << " " << std::setw (19) << l;
                // std::cout << " " << std::setw (19) << std::setprecision (precision) << val(0,0);
                // std::cout << " " << std::setw (19) << std::setprecision (precision) << val(0,1);
                // std::cout << " " << std::setw (19) << std::setprecision (precision) << val(1,0);
                // std::cout << " " << std::setw (19) << std::setprecision (precision) << val(1,1) << std::endl;
                EigenSystem esys;
                eigensystem_symm (&esys, val);
                if (fabs (esys.eval[0]) < fabs (esys.eval[1])) {
                    swap_eigenvalues (&esys);
                }
                double ratio = esys.eval[1]/esys.eval[0];
                output_ptr[l + 4*num_labels] = esys.eval[0];
                output_ptr[l + 5*num_labels] = esys.eval[1];
                output_ptr[l + 6*num_labels] = esys.evec[0][0];
                output_ptr[l + 7*num_labels] = esys.evec[0][1];
                output_ptr[l + 8*num_labels] = esys.evec[1][0];
                output_ptr[l + 9*num_labels] = esys.evec[1][1];
                // std::cout << "eigenvalues and eigenratio: " << std::endl;
                // std::cout << " " << std::setw (19) << std::setprecision (precision) << esys.eval[0];
                // std::cout << " " << std::setw (19) << std::setprecision (precision) << esys.eval[1];
                // std::cout << " " << std::setw (19) << std::setprecision (precision) << ratio << std::endl;
                // std::cout << "eigenvectors: " << std::endl;
                // std::cout << " " << std::setw (19) << std::setprecision (precision) << esys.evec[0][0];
                // std::cout << " " << std::setw (19) << std::setprecision (precision) << esys.evec[0][1];
                // std::cout << " " << std::setw (19) << std::setprecision (precision) << esys.evec[1][0];
                // std::cout << " " << std::setw (19) << std::setprecision (precision) << esys.evec[1][1] << std::endl;
            }
            break;
        }
        case 8: {
            std::cout << "w120:"  << std::endl;
            MatrixMinkowskiFunctional *p = dynamic_cast<MatrixMinkowskiFunctional*>(m_func);
            for (int l = 0; l != num_labels; ++l) {
                mat_t val = p->value (l);
                output_ptr[l] = val(0,0);
                output_ptr[l + num_labels] = val(0,1);
                output_ptr[l + 2*num_labels] = val(1,0);
                output_ptr[l + 3*num_labels] = val(1,1);
                // std::cout << " " << std::setw (19) << l;
                // std::cout << " " << std::setw (19) << std::setprecision (precision) << val(0,0);
                // std::cout << " " << std::setw (19) << std::setprecision (precision) << val(0,1);
                // std::cout << " " << std::setw (19) << std::setprecision (precision) << val(1,0);
                // std::cout << " " << std::setw (19) << std::setprecision (precision) << val(1,1) << std::endl;
                EigenSystem esys;
                eigensystem_symm (&esys, val);
                if (fabs (esys.eval[0]) < fabs (esys.eval[1])) {
                    swap_eigenvalues (&esys);
                }
                double ratio = esys.eval[1]/esys.eval[0];
                output_ptr[l + 4*num_labels] = esys.eval[0];
                output_ptr[l + 5*num_labels] = esys.eval[1];
                output_ptr[l + 6*num_labels] = esys.evec[0][0];
                output_ptr[l + 7*num_labels] = esys.evec[0][1];
                output_ptr[l + 8*num_labels] = esys.evec[1][0];
                output_ptr[l + 9*num_labels] = esys.evec[1][1];
                // std::cout << "eigenvalues and eigenratio: " << std::endl;
                // std::cout << " " << std::setw (19) << std::setprecision (precision) << esys.eval[0];
                // std::cout << " " << std::setw (19) << std::setprecision (precision) << esys.eval[1];
                // std::cout << " " << std::setw (19) << std::setprecision (precision) << ratio << std::endl;
                // std::cout << "eigenvectors: " << std::endl;
                // std::cout << " " << std::setw (19) << std::setprecision (precision) << esys.evec[0][0];
                // std::cout << " " << std::setw (19) << std::setprecision (precision) << esys.evec[0][1];
                // std::cout << " " << std::setw (19) << std::setprecision (precision) << esys.evec[1][0];
                // std::cout << " " << std::setw (19) << std::setprecision (precision) << esys.evec[1][1] << std::endl;
            }
            break;
        }
        case 9: {
            std::cout << "w102:"  << std::endl;
            MatrixMinkowskiFunctional *p = dynamic_cast<MatrixMinkowskiFunctional*>(m_func);
            for (int l = 0; l != num_labels; ++l) {
                mat_t val = p->value (l);
                output_ptr[l] = val(0,0);
                output_ptr[l + num_labels] = val(0,1);
                output_ptr[l + 2*num_labels] = val(1,0);
                output_ptr[l + 3*num_labels] = val(1,1);
                // std::cout << " " << std::setw (19) << l;
                // std::cout << " " << std::setw (19) << std::setprecision (precision) << val(0,0);
                // std::cout << " " << std::setw (19) << std::setprecision (precision) << val(0,1);
                // std::cout << " " << std::setw (19) << std::setprecision (precision) << val(1,0);
                // std::cout << " " << std::setw (19) << std::setprecision (precision) << val(1,1) << std::endl;
                EigenSystem esys;
                eigensystem_symm (&esys, val);
                if (fabs (esys.eval[0]) < fabs (esys.eval[1])) {
                    swap_eigenvalues (&esys);
                }
                double ratio = esys.eval[1]/esys.eval[0];
                output_ptr[l + 4*num_labels] = esys.eval[0];
                output_ptr[l + 5*num_labels] = esys.eval[1];
                output_ptr[l + 6*num_labels] = esys.evec[0][0];
                output_ptr[l + 7*num_labels] = esys.evec[0][1];
                output_ptr[l + 8*num_labels] = esys.evec[1][0];
                output_ptr[l + 9*num_labels] = esys.evec[1][1];
                // std::cout << "eigenvalues and eigenratio: " << std::endl;
                // std::cout << " " << std::setw (19) << std::setprecision (precision) << esys.eval[0];
                // std::cout << " " << std::setw (19) << std::setprecision (precision) << esys.eval[1];
                // std::cout << " " << std::setw (19) << std::setprecision (precision) << ratio << std::endl;
                // std::cout << "eigenvectors: " << std::endl;
                // std::cout << " " << std::setw (19) << std::setprecision (precision) << esys.evec[0][0];
                // std::cout << " " << std::setw (19) << std::setprecision (precision) << esys.evec[0][1];
                // std::cout << " " << std::setw (19) << std::setprecision (precision) << esys.evec[1][0];
                // std::cout << " " << std::setw (19) << std::setprecision (precision) << esys.evec[1][1] << std::endl;
            }
            break;
        }
        case 10: {
            std::cout << "w220:"  << std::endl;
            MatrixMinkowskiFunctional *p = dynamic_cast<MatrixMinkowskiFunctional*>(m_func);
            for (int l = 0; l != num_labels; ++l) {
                mat_t val = p->value (l);
                output_ptr[l] = val(0,0);
                output_ptr[l + num_labels] = val(0,1);
                output_ptr[l + 2*num_labels] = val(1,0);
                output_ptr[l + 3*num_labels] = val(1,1);
                // std::cout << " " << std::setw (19) << l;
                // std::cout << " " << std::setw (19) << std::setprecision (precision) << val(0,0);
                // std::cout << " " << std::setw (19) << std::setprecision (precision) << val(0,1);
                // std::cout << " " << std::setw (19) << std::setprecision (precision) << val(1,0);
                // std::cout << " " << std::setw (19) << std::setprecision (precision) << val(1,1) << std::endl;
                EigenSystem esys;
                eigensystem_symm (&esys, val);
                if (fabs (esys.eval[0]) < fabs (esys.eval[1])) {
                    swap_eigenvalues (&esys);
                }
                double ratio = esys.eval[1]/esys.eval[0];
                output_ptr[l + 4*num_labels] = esys.eval[0];
                output_ptr[l + 5*num_labels] = esys.eval[1];
                output_ptr[l + 6*num_labels] = esys.evec[0][0];
                output_ptr[l + 7*num_labels] = esys.evec[0][1];
                output_ptr[l + 8*num_labels] = esys.evec[1][0];
                output_ptr[l + 9*num_labels] = esys.evec[1][1];
                // std::cout << "eigenvalues and eigenratio: " << std::endl;
                // std::cout << " " << std::setw (19) << std::setprecision (precision) << esys.eval[0];
                // std::cout << " " << std::setw (19) << std::setprecision (precision) << esys.eval[1];
                // std::cout << " " << std::setw (19) << std::setprecision (precision) << ratio << std::endl;
                // std::cout << "eigenvectors: " << std::endl;
                // std::cout << " " << std::setw (19) << std::setprecision (precision) << esys.evec[0][0];
                // std::cout << " " << std::setw (19) << std::setprecision (precision) << esys.evec[0][1];
                // std::cout << " " << std::setw (19) << std::setprecision (precision) << esys.evec[1][0];
                // std::cout << " " << std::setw (19) << std::setprecision (precision) << esys.evec[1][1] << std::endl;
            }
            break;
        }
        case 11: {
            std::cout << "w211:"  << std::endl;
            MatrixMinkowskiFunctional *p = dynamic_cast<MatrixMinkowskiFunctional*>(m_func);
            for (int l = 0; l != num_labels; ++l) {
                mat_t val = p->value (l);
                output_ptr[l] = val(0,0);
                output_ptr[l + num_labels] = val(0,1);
                output_ptr[l + 2*num_labels] = val(1,0);
                output_ptr[l + 3*num_labels] = val(1,1);
                // std::cout << " " << std::setw (19) << l;
                // std::cout << " " << std::setw (19) << std::setprecision (precision) << val(0,0);
                // std::cout << " " << std::setw (19) << std::setprecision (precision) << val(0,1);
                // std::cout << " " << std::setw (19) << std::setprecision (precision) << val(1,0);
                // std::cout << " " << std::setw (19) << std::setprecision (precision) << val(1,1) << std::endl;
                EigenSystem esys;
                eigensystem_symm (&esys, val);
                if (fabs (esys.eval[0]) < fabs (esys.eval[1])) {
                    swap_eigenvalues (&esys);
                }
                double ratio = esys.eval[1]/esys.eval[0];
                output_ptr[l + 4*num_labels] = esys.eval[0];
                output_ptr[l + 5*num_labels] = esys.eval[1];
                output_ptr[l + 6*num_labels] = esys.evec[0][0];
                output_ptr[l + 7*num_labels] = esys.evec[0][1];
                output_ptr[l + 8*num_labels] = esys.evec[1][0];
                output_ptr[l + 9*num_labels] = esys.evec[1][1];
                // std::cout << "eigenvalues and eigenratio: " << std::endl;
                // std::cout << " " << std::setw (19) << std::setprecision (precision) << esys.eval[0];
                // std::cout << " " << std::setw (19) << std::setprecision (precision) << esys.eval[1];
                // std::cout << " " << std::setw (19) << std::setprecision (precision) << ratio << std::endl;
                // std::cout << "eigenvectors: " << std::endl;
                // std::cout << " " << std::setw (19) << std::setprecision (precision) << esys.evec[0][0];
                // std::cout << " " << std::setw (19) << std::setprecision (precision) << esys.evec[0][1];
                // std::cout << " " << std::setw (19) << std::setprecision (precision) << esys.evec[1][0];
                // std::cout << " " << std::setw (19) << std::setprecision (precision) << esys.evec[1][1] << std::endl;
            }
            break;
        }
    
    } 

    delete m_func;

    
}
