% calculate_ellipse_params.m

% function to convert the eigenvalues from the W020, W120, and W220
% Minkowski tensors into ellipse parameters
% according to Rahman & Shandarin, MNRAS 343 (2003)
% the eigenvectors of the tensors yield the orientation angle of the
% semi-major axis

% input: mink_type (7: W020, 8: W120, 10:W220) define tensor
%        mink_tensor (2 x 2 matrix of doubles) Minkowski tensor
%
% output: ell_params (1 x 6 double) 
%                     area, perimeter, orientation angle(in deg), 
%                     semi-major, semi-minor axis, minor/major axis ratio
% 

function [ell_params] = calculate_ellipse_params(mink_type, mink_tensor)

% define precision below eccentricity is considered zero:
eps = 1e-06;

% diagonalise Minkowski tensor (retrieve eigenvalues and eigenvectors)
[eigenvecs,eigenvals] = eig(mink_tensor);

% solve for ellipse axis parameters (a, b or e = sqrt(1-b^2/(a^2))
switch mink_type
    
    case 7 % W020
        axis1 = (eigenvals(1,1)*pi*((16*eigenvals(2,2)^3)/(eigenvals(1,1)*pi^2))^(5/8))/(4*eigenvals(2,2)^2);
        axis2 = ((16*eigenvals(2,2)^3)/(eigenvals(1,1)*pi^2))^(1/8);
        
    case 8 % W120 (could be accelerated)
        
        sol = fsolve(@(var)ellipse_equations_W120(var,eigenvals),[10,0.5]);
        
        % consistency check (non-unique solution cannot occur)
        if(abs(sol(2)) < eps)
            warning(['eccentricity ', num2str(sol(2)), ' is zero.']);
            sol(2) = 0;
        else
            if(sol(2) < 0)
                error(['eccentricity ', num2str(sol(2)), ' negative.']);
            end
        end
        
        if(sol(2) > 1)
            error(['eccentricity ', num2str(sol(2)), ' larger than 1.']);
        end
        
        if (abs(imag(sol(1))) > eps )
            error(['solution for a ', num2str(sol(1)), ' is complex.']);
        end
        
        if(abs(imag(sol(2))) > eps )
            error(['solution for e ', num2str(sol(2)), ' is complex.']);
        end
        
        axis1 = double(sol(1));
        axis2 = double(axis1*sqrt(1-sol(2)));
        
    case 10 % W220
        syms a; 
        syms b;
        assume(a,'real'); 
        assumeAlso(a,'positive');
        assume(b,'real'); 
        assumeAlso(b,'positive');
        sol = solve([eigenvals(1,1) == b^3/(a+b), eigenvals(2,2) == a^3/(a+b)], [a,b]);
        
        if(size(sol,1) > 1)
            error('solution for W220-tensor not unique.');
        end
        
        axis1 = double(sol.a);
        axis2 = double(sol.b);
        
    otherwise
        error(['Minkowski tensor ', num2str(mink_type), ' not yet implemented.']);
        
end

%% calculate orientation angle of semi-major axis
if(eigenvals(1,1) > eigenvals(2,2))
    semi_major = 1;
    
else
    semi_major = 2;
end

% calculate semi-major and semi-minor ellipse axis
if(axis1 > axis2)
    major = axis1;
    minor = axis2;
else
    major = axis2;
    minor = axis1;
end
    
syms varphi;

ell_params(1,1) = pi*axis1*axis2;
ell_params(1,2) = 4*major*double(int(sqrt(1-(1-minor^2/(major^2))*(sin(varphi))^2),0,pi/2));
% orientation angle of semi-major axis in range [-90, +90]
ell_params(1,3) = atan(eigenvecs(2,semi_major)/eigenvecs(1,semi_major))*180/pi;
ell_params(1,4) = major;
ell_params(1,5) = minor;
ell_params(1,6) = minor/major;
