%%%% example to show functionality of the MATLAB-Papaya-Interface
%%%% bug reports, questions and comments to 
%%%% Jenny Wagner (j.wagner@uni-heidelberg.de)

%%%% read in '../testdata/test_image.fits', 
%%%% specify parameters to calculate all (tensorial) Minkowski functionals 
%%%%   implemented in Papaya for the two regions of interest, including the
%%%%   segmentation of the image, by component,
%%%% calculate p_i-vectors as an example to detect deviations from 
%%%%                    elliptical symmetry in the two regions of interest
%%%% visualise results

clear all;

%%% specify parameters 
%% image to be read in
test_picture = '../testdata/test_image.fits';

%% parameters to calculate Minkowski functionals 
%% (passed to calculate_MF in exactly this order, path to output optional)
% name: 1: w000, 2: w100, 3: w200 (scalars)
%       4: w010, 5: w110, 6: w210 (vectors)
%       7: w020, 8: w120, 9: w102, 10: w220, 11: w211 (tensors)
% origin: 1: origin of picture
%         2: centre of mass
%         3: centre of perimeter
%         4: centre of curvature
% method: 1: by contour
%         2: by component
% threshold: double between 0.0 and 1.0 (no thresholding: set to 0.0)
% booleans to invert image, connect black, periodic data (all "false")
% precision: integer number of digits for the output
% write out labels if dump_labels is set to 1
% if labels are to be written out, specify path to output file 
% (called <path>/labels.out and <path>/labels.gp
mf_name = 1; % not required further below, just for demonstration here
mf_origin = 2; % not required further below, just for demonstration here
mf_method = 2;
threshold = 1.3151169*10^(-3);
invert_image = logical(0);
connect_black = logical(0);
periodic_data = logical(0);
precision = 15;
dump_labels = logical(1);
label_path = '../results/';

%% read in image, convert to double and normalise it
picture = fitsread(test_picture);
picture = double(picture)./max(max(double(picture)));

%% calculate Minkowski-functionals by contour for each region of interest
%% using calculate_MF-C++-function (MATLAB-Papaya-Interface)
% first input argument: mf_name
% second input argument: mf_origin    

tmp_picture = double((picture));
  
% define filename for labels
[label_filename] = sprintf('%s/combined_labels',label_path);

% scalars (for both ROIs in one vector):
[w000] = calculate_MF(1,1,mf_method,threshold,invert_image,connect_black,periodic_data,tmp_picture,precision,dump_labels,label_filename);
[w100] = calculate_MF(2,1,mf_method,threshold,invert_image,connect_black,periodic_data,tmp_picture,precision,0);
[w200] = calculate_MF(3,1,mf_method,threshold,invert_image,connect_black,periodic_data,tmp_picture,precision,0);

% vectors (for both ROIS in one matrix):
[w010] = calculate_MF(4,1,1,threshold,invert_image,connect_black,periodic_data,tmp_picture,precision,0);
[w110] = calculate_MF(5,1,1,threshold,invert_image,connect_black,periodic_data,tmp_picture,precision,0);
[w210] = calculate_MF(6,1,1,threshold,invert_image,connect_black,periodic_data,tmp_picture,precision,0);

% calculate p_i for each ROI
for cnt_roi=1:1:size(w000,1)
    p0(cnt_roi,:) = w010(cnt_roi,:)./w000(cnt_roi);
    p1(cnt_roi,:) = w110(cnt_roi,:)./w100(cnt_roi);
    p2(cnt_roi,:) = w210(cnt_roi,:)./w200(cnt_roi);
end

% tensors (for both ROIS in a struct):
[w020] = calculate_MF(7,2,mf_method,threshold,invert_image,connect_black,periodic_data,tmp_picture,precision,0);
[w120] = calculate_MF(8,3,mf_method,threshold,invert_image,connect_black,periodic_data,tmp_picture,precision,0);
[w220] = calculate_MF(10,4,mf_method,threshold,invert_image,connect_black,periodic_data,tmp_picture,precision,0);

%% visualise results
% show image
figure; subplot(1,2,1);
imagesc(log(picture)); 
% generate mask for labels to be overlaid with segmented image
label_mask = zeros(size(picture)); 
% generate mask for p_is to be overlaid with segmented image
p0_label_mask = zeros(size(picture));
p1_label_mask = zeros(size(picture));
p2_label_mask = zeros(size(picture));
% generate mask for ellipse defined by eigenvalues of w020
w020_label_mask = zeros([size(picture),2]);

% read in labels
[label_filename] = sprintf('%s/combined_labels.out',label_path);
labels = load(label_filename);
for cnt_roi=1:1:size(w000,1)
    
    % set mask to 1 at label positions
    for cnt_labels=1:1:size(labels,1)
        label_mask(uint16(labels(cnt_labels,2)),uint16(labels(cnt_labels,1))) = 1;
    end
    
    % set mask to 1 at p_i positions
    p0_label_mask(uint16(p0(cnt_roi,2)),uint16(p0(cnt_roi,1))) = 1;
    p1_label_mask(uint16(p1(cnt_roi,2)),uint16(p1(cnt_roi,1))) = 1;
    p2_label_mask(uint16(p2(cnt_roi,2)),uint16(p2(cnt_roi,1))) = 1;
    
    % determine ellipse of w020
    % according to Rahman & Shandarin, MNRAS 343 (2003)
    ell_height = (w020(cnt_roi,6)*pi*((16*w020(cnt_roi,5)^3)/(w020(cnt_roi,6)*pi^2))^(5/8))/(4*w020(cnt_roi,5)^2);
    ell_width = ((16*w020(cnt_roi,5)^3)/(w020(cnt_roi,6)*pi^2))^(1/8);
    ell_xmin = p0(cnt_roi,1) - ell_width;
    ell_ymin = p0(cnt_roi,2) - ell_height;
    
    % set mask to 1 at ellipse positions
    h = imellipse(gca, [ell_xmin ell_ymin 2*ell_width 2*ell_height]);
    w020_label_mask(:,:,cnt_roi) = h.createMask();
end

% generate overlays:
lab_picture = bwlabel(picture > threshold); 
% labels in red
labelled_image = imoverlay(lab_picture,logical(label_mask),[1 0 0]);
% centre of mass point in blue
p0_labelled_image = imoverlay(labelled_image,logical(p0_label_mask),[0 0 1]);
% centre of perimeter in green
p1_labelled_image = imoverlay(p0_labelled_image,logical(p1_label_mask),[0 1 0]);
% centre of curvature in purple
p2_labelled_image = imoverlay(p1_labelled_image,logical(p2_label_mask),[1 0 1]);
% w020-ellipses in yellow
w1_labelled_image = imoverlay(p2_labelled_image,logical(bwperim(w020_label_mask(:,:,1))),[1 1 0]);
w2_labelled_image = imoverlay(w1_labelled_image,logical(bwperim(w020_label_mask(:,:,2))),[1 1 0]);
% plot overlaid segmented image detail
subplot(1,2,2); image(w2_labelled_image(1010:1040,720:900,:)); 

