%%%% example to show functionality of the MATLAB-Papaya-Interface
%%%% bug reports, questions and comments to 
%%%% Jenny Wagner (j.wagner@uni-heidelberg.de)

%%%% read in '../testdata/slika.pgm', segment it into different patches,
%%%% calculate w000, w100, and w200 for each patch and visualise the
%%%% results as three images with pixelvalues representing w000, w100, and
%%%% w200

clear all;

%%% specify parameters 
%% image to be read in
test_picture = '../testdata/slika.pgm';

%% parameters to calculate Minkowski functionals
%% (passed to calculate_MF in exactly this order, path to output optional)
% name: 1: w000, 2: w100, 3: w200 (scalars)
%       4: w010, 5: w110, 6: w210 (vectors)
%       7: w020, 8: w120, 9: w102, 10: w220, 11: w211 (tensors)
% origin: 1: origin of picture
%         2: centre of mass
%         3: centre of perimeter
%         4: centre of curvature
% method: 1: by contour
%         2: by component
% threshold: double between 0.0 and 1.0 (no thresholding: set to 0.0)
% booleans to invert image, connect black, periodic data (all "false")
% precision: integer number of digits for the output
% write out labels if dump_labels is set to 1
% if labels are to be written out, specify path to output file 
% (called <label_path>/labels.out and <label_path>/labels.gp
mf_name = 1; % not required further below, just for demonstration here
mf_origin = 2; % not required further below, just for demonstration here
mf_method = 1;
threshold = 0.0; % image is already segmented
invert_image = logical(0);
connect_black = logical(0);
periodic_data = logical(0);
precision = 15;
dump_labels = logical(0);

%% read in image, convert to double and normalise it
picture = imread(test_picture);
%picture = imread(test_picture); 
picture = double(picture)./max(max(double(picture)));

%% split image into patches
patchsize_x = 47;
patchsize_y = 104;
x = 1:patchsize_x:size(picture,1)-patchsize_x+1;
y = 1:patchsize_y:size(picture,2)-patchsize_y+1;

%% calculate scalar MFs for each patch
area = zeros(size(x,1),size(y,1));
perimeter = zeros(size(x,1),size(y,1));
curvature = zeros(size(x,1),size(y,1));

for cnt_x=1:1:size(x,2)
    for cnt_y=1:1:size(y,2)
        
        % cut out patch of picture
        tmp_picture = picture(x(cnt_x):x(cnt_x)+patchsize_x-1,y(cnt_y):y(cnt_y)+patchsize_y-1);
        % calculate MFs
        [w000] = calculate_MF(1,1,mf_method,threshold,invert_image,connect_black,periodic_data,tmp_picture,precision,0);
        [w100] = calculate_MF(2,1,mf_method,threshold,invert_image,connect_black,periodic_data,tmp_picture,precision,0);
        [w200] = calculate_MF(3,1,mf_method,threshold,invert_image,connect_black,periodic_data,tmp_picture,precision,0);
        % add all contours to one MF for each w, due to additivity of MFs
        area(cnt_x,cnt_y) = sum(w000); 
        perimeter(cnt_x,cnt_y) = sum(w100);
        curvature(cnt_x,cnt_y) = sum(w200);

    end
end

%% visualise results
% initialise picture with patches (draw red lines for patch boundaries)
patch_mask = zeros(size(picture));
patch_mask(x,:) = 1;
patch_mask(:,y) = 1;
patch_mask(x+1,:) = 1;
patch_mask(:,y+1) = 1;
% plot results
figure; 
subplot(2,2,1); imagesc(imoverlay(picture,logical(patch_mask),[1 0 0])); title('patched picture');
subplot(2,2,2); imagesc(area); colorbar; title('area per patch');
subplot(2,2,3); imagesc(perimeter); colorbar; title('perimeter per patch');
subplot(2,2,4); imagesc(curvature); colorbar; title('curvature per patch');
