%%%% example to show functionality of the MATLAB-Papaya-Interface
%%%% bug reports, questions and comments to 
%%%% Jenny Wagner (j.wagner@uni-heidelberg.de)

%%%% read in '../testdata/ellipse.pgm' or '../testdata/circle.pgm'
%%%% specify parameters to calculate all (tensorial) Minkowski functionals 
%%%%     implemented in Papaya for the two regions of interest, by contour,
%%%% calculate p_i-vectors as an example to detect deviations from 
%%%%                    elliptical symmetry in the two regions of interest
%%%% calculate W020, W120, W220 ellipse parameters acc. to Rahman &
%%%% Shandarin, MNRAS 343 (2003)
%%%% visualise results including overlay with labels

clear all;

%%% specify parameters 
%% image to be read in
%test_picture = '../testdata/circle.pgm'
test_picture = '../testdata/ellipse.pgm'

%% parameters to calculate Minkowski functionals
%% (passed to calculate_MF in exactly this order, path to output optional)
% name: 1: w000, 2: w100, 3: w200 (scalars)
%       4: w010, 5: w110, 6: w210 (vectors)
%       7: w020, 8: w120, 9: w102, 10: w220, 11: w211 (tensors)
% origin: 1: origin of picture
%         2: centre of mass
%         3: centre of perimeter
%         4: centre of curvature
% method: 1: by contour
%         2: by component
% threshold: double between 0.0 and 1.0 (no thresholding: set to 0.0)
% booleans to invert image, connect black, periodic data (all "false")
% precision: integer number of digits for the output
% write out labels if dump_labels is set to 1
% if labels are to be written out, specify path to output file 
% (called <label_path>/labels.out and <label_path>/labels.gp
mf_name = 1; % not required further below, just for demonstration here
mf_origin = 2; % not required further below, just for demonstration here
mf_method = 1;
threshold = 0.0;
invert_image = logical(0);
connect_black = logical(0);
periodic_data = logical(0);
precision = 15;
dump_labels = logical(1);
label_path = '../results';

%% read in image, convert to double and normalise it
picture = imread(test_picture); 
picture = double(picture)./max(max(double(picture)));

%% find the two regions of interest
seg_picture = picture;
lab_picture = bwlabel(seg_picture);

%% find number of regions of interest (yields two regions of interest)
num_labels = max(max(lab_picture)); 

%% calculate Minkowski-functionals by contour for each region of interest
%% using calculate_MF-C++-function (MATLAB-Papaya-Interface)
% first input argument: mf_name
% second input argument: mf_origin
for cnt_roi=1:1:num_labels
    
    tmp_picture = double((lab_picture == cnt_roi));
    
    % define filename for labels for each ROI
    [label_filename] = sprintf('%s/labels_%d',label_path, cnt_roi);
    
    % scalars (for both ROIs in one vector):
    [w000(cnt_roi)] = calculate_MF(1,1,mf_method,threshold,invert_image,connect_black,periodic_data,tmp_picture,precision,dump_labels,label_filename);
    [w100(cnt_roi)] = calculate_MF(2,1,mf_method,threshold,invert_image,connect_black,periodic_data,tmp_picture,precision,0);
    [w200(cnt_roi)] = calculate_MF(3,1,mf_method,threshold,invert_image,connect_black,periodic_data,tmp_picture,precision,0);
    
    % vectors (results for each entry in a row):
    % first entry: x-coordinate, second entry: y-coordinate
    [w010(cnt_roi,:)] = calculate_MF(4,1,1,threshold,invert_image,connect_black,periodic_data,tmp_picture,precision,0);
    [w110(cnt_roi,:)] = calculate_MF(5,1,1,threshold,invert_image,connect_black,periodic_data,tmp_picture,precision,0);
    [w210(cnt_roi,:)] = calculate_MF(6,1,1,threshold,invert_image,connect_black,periodic_data,tmp_picture,precision,0);
    
    % calculate p_i
    p0(cnt_roi,:) = w010(cnt_roi,:)./w000(cnt_roi);
    p1(cnt_roi,:) = w110(cnt_roi,:)./w100(cnt_roi);
    p2(cnt_roi,:) = w210(cnt_roi,:)./w200(cnt_roi);
    
    % tensors (results for each ROI in a line):
    % first four entries: a11, a12, a21, a22
    % next two entries: largest and second eigenvalue
    % next two entries: eigenvector (x,y) of largest eigenvalue
    % next two entries: eigenvector (x,y) of second eigenvalue
    [w020(cnt_roi,:)] = calculate_MF(7,2,mf_method,threshold,invert_image,connect_black,periodic_data,tmp_picture,precision,0);
    [w120(cnt_roi,:)] = calculate_MF(8,3,mf_method,threshold,invert_image,connect_black,periodic_data,tmp_picture,precision,0);
    [w220(cnt_roi,:)] = calculate_MF(10,4,mf_method,threshold,invert_image,connect_black,periodic_data,tmp_picture,precision,0);
    
    
end

%% visualise results
% show image
figure; subplot(1,2,1);
imagesc(log(picture)); 
% generate mask for labels to be overlaid with segmented image
label_mask = zeros(size(picture)); 
% generate mask for p_is to be overlaid with segmented image
p0_label_mask = zeros(size(picture));
p1_label_mask = zeros(size(picture));
p2_label_mask = zeros(size(picture));
% generate mask for ellipses defined by eigenvalues of w020, w120, w220
w020_label_mask = zeros([size(picture),2]);
w120_label_mask = zeros([size(picture),2]);
w220_label_mask = zeros([size(picture),2]);

for cnt_roi=1:1:num_labels
    
    % read in labels
    [label_filename] = sprintf('%s/labels_%d.out',label_path, cnt_roi);
    labels = load(label_filename);
    
    % set mask to 1 at label positions
    for cnt_labels=1:1:size(labels,1)
                label_mask((uint16(labels(cnt_labels,2))),(uint16(labels(cnt_labels,1)))) = 1;
    end
    
    % set mask to 1 at p_i positions (and shifted by 1px in y-direction)
    p0_label_mask(uint16(p0(cnt_roi,2)),uint16(p0(cnt_roi,1))) = 1;
    p1_label_mask(uint16(p1(cnt_roi,2)),uint16(p1(cnt_roi,1))) = 1;
    p2_label_mask(uint16(p2(cnt_roi,2)),uint16(p2(cnt_roi,1))) = 1;
    
    % determine ellipse of w120
    [w120_ell_params(cnt_roi,:)] = calculate_ellipse_params(8, [w120(cnt_roi,1), w120(cnt_roi,2); w120(cnt_roi,3), w120(cnt_roi,4)]);
    ell_xmin = p1(cnt_roi,1) - w120_ell_params(cnt_roi,4);
    ell_ymin = p1(cnt_roi,2)+1 - w120_ell_params(cnt_roi,5);
    % set mask to 1 at ellipse positions
    h = imellipse(gca, [ell_xmin ell_ymin 2*w120_ell_params(cnt_roi,4) 2*w120_ell_params(cnt_roi,5)]);
    w120_label_mask(:,:,cnt_roi) = h.createMask();
    
    % determine ellipse of w220
    [w220_ell_params(cnt_roi,:)] = calculate_ellipse_params(10, 1/(2*pi).*[w220(cnt_roi,1), w220(cnt_roi,2); w220(cnt_roi,3), w220(cnt_roi,4)]);
    ell_xmin = p2(cnt_roi,1) - w220_ell_params(cnt_roi,4);
    ell_ymin = p2(cnt_roi,2)+1 - w220_ell_params(cnt_roi,5);
    % set mask to 1 at ellipse positions
    h = imellipse(gca, [ell_xmin ell_ymin 2*w220_ell_params(cnt_roi,4) 2*w220_ell_params(cnt_roi,5)]);
    w220_label_mask(:,:,cnt_roi) = h.createMask();
    
    % determine ellipse of w020
    [w020_ell_params(cnt_roi,:)] = calculate_ellipse_params(7, [w020(cnt_roi,1), w020(cnt_roi,2); w020(cnt_roi,3), w020(cnt_roi,4)]);
    ell_xmin = p0(cnt_roi,1) - w020_ell_params(cnt_roi,4);
    ell_ymin = p0(cnt_roi,2)+1 - w020_ell_params(cnt_roi,5);
    % set mask to 1 at ellipse positions
    h = imellipse(gca, [ell_xmin ell_ymin 2*w020_ell_params(cnt_roi,4) 2*w020_ell_params(cnt_roi,5)]);
    w020_label_mask(:,:,cnt_roi) = h.createMask();
    
end

% generate overlays:
% labels in yellow
labelled_image = imoverlay(lab_picture,logical(label_mask),[1 1 0]);
% centre of mass point in blue
p0_labelled_image = imoverlay(labelled_image,logical(p0_label_mask),[0 0 1]);
% centre of perimeter in green
p1_labelled_image = imoverlay(p0_labelled_image,logical(p1_label_mask),[0 1 0]);
% centre of curvature in red
p2_labelled_image = imoverlay(p1_labelled_image,logical(p2_label_mask),[1 0 0]);
% w020-ellipses in blue
w1_labelled_image = imoverlay(p2_labelled_image,logical(bwperim(w020_label_mask(:,:,1))),[0 0 1]);
w2_labelled_image = imoverlay(w1_labelled_image,logical(bwperim(w020_label_mask(:,:,2))),[0 0 1]);
% w120-ellipses in green
w3_labelled_image = imoverlay(w2_labelled_image,logical(bwperim(w120_label_mask(:,:,1))),[0 1 0]);
w4_labelled_image = imoverlay(w3_labelled_image,logical(bwperim(w120_label_mask(:,:,2))),[0 1 0]);
% w220-ellipses in red
w5_labelled_image = imoverlay(w4_labelled_image,logical(bwperim(w220_label_mask(:,:,1))),[1 0 0]);
w6_labelled_image = imoverlay(w5_labelled_image,logical(bwperim(w220_label_mask(:,:,2))),[1 0 0]);

% plot overlaid segmented image detail
subplot(1,2,2); image(w6_labelled_image(850:1200,750:1280,:)); 
